=begin
###
Copyright 2012, TIG (c)
###
Permission to use, copy, modify, and distribute this software for 
any purpose and without fee is hereby granted, provided the above
copyright notice appear in all copies.
THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
###
Description:
A toolset to 'Slice' up a volume - useful in real-model 
making or finding likely areas in master plan volumes...
###		  
Activation:
This version loads as an Extension and may initially require activating 
from the Window > Preferences > Extension dialog [=Slicer5].
Activate the Toolbar from View > Toolbars [=Slicer5].
###
Menu:
  Plugins > Slicer5
###
Toolbar:
  View > Toolbars > Slicer5
###
Context-Menu:
  N/A
###
Ruby Console:
  TIG::Slicer.new()
###
Usage:
This tool is for Sketchup version >= v8 only.
The DXF option is only available to v8-Pro users.
It requires the Sliced Object to be a Manifold Solid.
Select ONE suitable Group or Component-Instance and run 'Slicer5'.
A dialog asks for the Slicer5 Parameters [these settings are remembered 
and reused during the session until they are next changed].

Axis:
X
Y
Z
XY
XZ
YZ
- 'Z' is the default
The 'axis' is the RGB/XYZ axis along which the Slices will be taken.
If you choose a dual axis [XY/XZ/YZ] then Slices will be taken on those 
two axes and additional dialogs will appear - see later...
Select the desires axis from the list.

Spacing:
- default is 2" or 50mm
Sets the center-to-center Spacing of the Slices.
The slices are set-out centered on the Sliced Object's center.
The first and last Slices Spacings are automatically adjusted to fit 
within the bounds of the Sliced Object and any Insets [see below].
The typed in value assumes current-units, you can add a suffix to use an 
alternative - e.g. if you are working in 'cm', typing '5' gives 5cm, or 
typing '50mm' gives 5cm too.
If you enter 0 it assumes you want to set a total number of 'spacings', 
with the actual spacing dimension calculated and adjusted to suit - 
a second dialog will ask for details of this later [see below]...

Thickness:
- default is 1/8" or 3mm
Sets the slice Thickness, it cannot be more than the Spacing.
The typed in value assumes current-units, you can add a suffix to use an 
alternative - e.g. if you are working in 'cm', typing '5' gives 5cm, or 
typing '50mm' gives 5cm too.

Inset at Start:
- default 0
Sets the first Slice's inset, 0=flush with Sliced Object's bounds.

Inset at End:
- default 0
Sets the last Slice's inset, 0=flush with Sliced Object's bounds.

Add References ?
- default 'Yes'
Adds a 'TEXT' reference tag to the bottom-left corner of each Slice.
It might not always be on a face, but can be relocated manually as 
desired.
If the system-font 'Txt.ttf' is installed, then it will be used to make 
a 'stick' font flat-text group [TEXT] on layer 'SLCR-TEXT'; otherwise 
the current 3d_Text font will be used.
The Text Height is set in the next option...

Text Height: [from v5.3]
- default 0.5" or 12mm
Sets the Text Height of the Reference 'TEXT' [if specified].

Flatten ?
- default 'Yes'
This makes a flattened set of the Slices.
It is a group located to the max-extents of the Sliced Object.

'Cancel' stops the tool
'OK' runs the tool.

Since v5.2 if you chose a dual axis [XY/XZ/YZ] then a second dialog asks 
you for further options:

Spacing 2:
- generally as the Spacing above...
It defaults to that Spacing, but allows you to specify a different 
Spacing value for the second set of Slices.

Thickness2:
- generally as the Thickness above...
It defaults to that Thickness, but allows you to specify a different 
Thickness value for the second set of Slices.

Inset at Start2:
- generally as the Inset at Start above...
- default Inset at Start
Sets the first Slice2's inset, 0=flush with Sliced Object's bounds.

Inset at End2:
- generally as the Inset at End above...
- default Inset at Start
Sets the last Slice2's inset, 0=flush with Sliced Object's bounds.

Slot ?
- default is 'Yes' [but reverts to 'No' if Thickness==0].
If 'Yes' this will 'Slot' the Intersecting Slices.
BUT if either of the Slices' Thickness==0 then there will be no 
Slotting done, and the value of 'Slotting' defaults to 'No'.
Note that adding 'Slots' greatly increases the processing time needed, 
but is often necessary is a 3d assembly is envisaged...

Tolerance:
If you are Slotting slices and using a CNC router it's advisable to 
make the Slots slightly bigger than the Thickness - typically 0.2mm.
Half of the Tolerance is shaved off each of the sides of each Slot.  
The total width of each Slot is thereby made bigger than its mating 
Slots' Thickness by that additional Tolerance.
The default is 0.
Enter the required value.
If Slotting is set to 'No', Tolerance is auto-set to 0.

Overcut:
If you are Slotting slices and using a CNC router it's advisable to 
make the Slots slightly longer so that the internal rounded corners left 
by the router-bit do not foul the mating Slice's Slot ends.
This is sometimes called 'dog-boning' or 'mouse-ears'.
Typically you set this to be the router-bit diameter - say 3/4/5/6mm
[it'll be made half of this, as the router-bit's radius is always used]
The default is 0.
Enter the required value.
It should never be greater than the Thickness, because a router-bit of 
that diameter wouldn't cut the Slot correctly anyway!
If 'Slotting' is set to 'No', Overcut is auto-set to 0.

'Cancel' stops the tool
'OK' runs the tool.

Since v5.2 f you have specified a Spacing of '0' for an axis then 
another dialog opens asking you for addition values for the 'Spaces'...

- default is 1 - Note that 1 is the minimum number of spaces in a Sliced 
set - producing a start and an end Slice only.
The prompt shows the axis and the total size in current-units - e.g.
'Spaces [X 1234mm]:'
Enter the number of 'Spaces' required on that axis.
For example, if you want the Slices to be around 50mm center-to-center 
then do a quick calculation - 1234/50=24.68, so use 24 or 25 spaces 
depending on those Slices' Thickness and your preferred 'gaps'...
If you simply want a central Slice type in 2 Spaces, or for 2 internal 
Slices use 3 Spaces etc...
If you have specified a dual axis [XY/XZ/YZ] AND have entered a Spacing 
of 0 for both axes, then there will be a second option added to the 
dialog, e.g. for an XY axis set:
'Spaces [X 1234mm]:'
'Spaces [Y 5678mm]:'
Enter the required number of 'Spaces' required for both axes.

'Cancel' stops the tool
'OK' runs the tool.

'Slicer5' now runs...

It reports its progress in the Status Bar...

The Slice-set is given a unique random 4 digit code - e.g. '1234'
The Slices are assembled inside a group named 'SLICES-1234', placed 
exactly over the Sliced Object, at the specified Spacing etc.
This group is made on its own layer:
'SLCR-ZZZZ' for Z-axis Slices
'SLCR-YYYY' for Y-axis Slices
'SLCR-XXXX' for X-axis Slices
If specified the individual Slice-components have reference tag-groups 
added, named 'TEXT', reading 'Z001' to 'Z999', the prefix varying with 
the specified axis [XYZ]. This TEXT is put on layer 'SLCR-TEXT'.
The Slice-components themselves are named to match their reference with 
the Slice-set's code as a prefix - e.g. '1234-Z001'.
The Slices are colored to match their axis - Red/Green/Blue==X/Y/Z.
The materials are called, respectively, 'SLCR-XXXX', 'SLCR-YYYY' and 
'SLCR-ZZZZ'.
If a Flattened-set is specified then an instance of each Slice-component 
is placed flat, to the top-right of the Sliced Object's maximum bounds, 
arranged in 'as square a pattern as possible'; each one spaced from its 
neighbor by a minimum of the Slices' Thickness.  
The Flattened-set is a group named 'FLAT-1234', on layer 'SLCR-FLAT'.  
Because the assembled and flat Slices are instances of the 
same Slice-component, then editing one edits the other...
If you have chosen a dual axis [XY/XZ/YZ] then two separate assemblies 
of Slices are made - the second one is named 'SLICES2-1234', and if 
there is a Flattened-set, then the separate second one is named 
'FLAT2-1234' and it is place 'above' the first one along the Y-axis.
On completion the Sliced Object's layer is set to 'SLCR-HIDN', which is
switched 'off' so that you can see the assembled Slice-set.

Export:
If you have made a Flattened-set [or sets] you are then asked if you 
want to make a folder of Images for each of the Slices.  
'Cancel' to skip this.
'OK' to make them.
The PNG's folder is made in the SKP's folder and is named 
'ModelName-Slicer5-PNGs-1234'.
The PNG files are named after the Slice-component thus: '1234-Z001.png'
If you are an 8-Pro user you will also be asked if you want to make DXF 
files for each Slice - useful for CNC processing etc...
'Cancel' to skip this.
'OK' to make them.
The DXF's folder is made in the SKP's folder and it is named 
'ModelName-Slicer5-DXFs-1234'.
The DXF files are named after the Slice-component, thus: '1234-Z001.dxf'
All loops are welded into a 'polyline', no faces are exported.
It uses the current DXF options set for Sketchup [edit these to suit 
before Slicing] - for example many CNC machines like 'R12' and 'mm'.
The following layer/color conventions are followed:
EXTL = Outer loops [yellow]
INTL = Inner loops - if any [green]
TEXT = Text - if any [red]
External and Internal loops are put on separate layers because the 
router-bit needs swap from outside to inside when cutting these forms.
The Text layer polylines can either be ignored or used to engrave a 
reference tag, say 1mm deep into the board.
Because the DXF reprocessing may continue after the main process stops 
it is advisable NOT to close the SKP or try to open any of the DXFs for 
several moments after completion.

The 'Slicing' is one step undo-able.

###

Slicer5 Exporter:

From v5.1 you can now Export the Slices later on.
this is useful if you want to move the Text-tags [quite likely when 
Slicing complex shapes] or even edit some Slices' geometry and so on...
###
Menu:
  Plugins > Slicer5 Exporter
###
Toolbar:
  View > Toolbars > Slicer5 Exporter
###
Context-Menu:
  N/A
###
Ruby Console:
  TIG::Slicer.export() [with a 'FLAT-1234' group already preselected]
###
Usage:
After making the Slices [with a Flattened version]; later on, perhaps 
after you have relocated some Text-tags or edited some Slices' geometry, 
you can then select the 'FLAT-1234' group & run the 'Slicer5 Exporter'.
It runs exactly as described in the 'Export' details above, except that 
the additional word 'Exporter' appears in its dialogs and the Status-
Bar text.
If there is a corresponding 'FLAT2.1234' group then that is also 
processed: you should NOT select it - select the 'FLAT.1234' group only 
- and if a 'FLAT2' version exists then it will be found automatically.

###
Donations: via PayPal.com to info @ revitrev.org
###
Version:	
5.0 20111230 First issue in new format.
5.1 20111231 Slicer5 Exporter tool added.
5.2 20120101 New Spacing2/Thickness2/Insets'2/Tolerance dialog added, 
			 Slot and Overcut [dogboning] added to the same new dialog.
			 Entering Spacing=0 now opens a dialog to enter the number 
			 of Spaces required, rather than using a fixed Spacing.
5.3 20120103 New 'Text-Height' option added to first dialog.
			 Recoded to avoid fails and small 'snags' with complex 
			 slotting solutions.
5.4 20120103 Stupid typo fixed in code set.
5.5 20120104 Spacing=0 glitch when reverting to 'length' fixed.
###
=end
###
require 'sketchup.rb'
###

module TIG

### menu ----------------------------------
unless file_loaded?(File.basename(__FILE__))
  cmd=UI::Command.new("Slicer5"){self::Slicer.new()}
  cmd.tooltip="Slicer5"
  cmd.status_bar_text="Slicer5: Slicing tool..."
  cmd.small_icon="Slicer5-16.png"
  cmd.large_icon="Slicer5-24.png"
  ###
  cmdx=UI::Command.new("Slicer5 Exporter"){self::Slicer.export()}
  cmdx.tooltip="Slicer5 Exporter"
  cmdx.status_bar_text="Slicer5: Exporter tool..."
  cmdx.small_icon="Slicer5-exporter-16.png"
  cmdx.large_icon="Slicer5-exporter-24.png"
  ###
  UI.menu("Plugins").add_item(cmd)
  UI.menu("Plugins").add_item(cmdx)
  ###
  toolbar=UI::Toolbar.new("Slicer5")
  toolbar.add_item(cmd)
  toolbar.add_item(cmdx)
  toolbar.restore if toolbar.get_last_state.abs==1 #TB_VISIBLE/NEVER
end#if
file_loaded(File.basename(__FILE__))

class self::Slicer
###---------------------------------------
   ###
   @@axis=nil
   @@spacing1=nil
   @@thickness1=nil
   @@inset_start1=nil
   @@inset_end1=nil
   ###
   @@ref=nil
   @@tht=nil
   @@flatten=nil
   ###
   @@spacing2=nil
   @@thickness2=nil
   @@inset_start2=nil
   @@inset_end2=nil
   ###
   @@slot=nil
   @@tol=nil
   @@bit=nil
   ###
   @@spaces1=0
   @@spaces2=0
   ###
###---------------------------------------

def initialize()
  if Sketchup.version.split(".")[0].to_i < 8
    UI.messagebox("Slicer5:\n\nThis Tool for Version 8+ ONLY !")
	return nil
  end#if
  @model=Sketchup.active_model
  if @model.path.empty?
    UI.messagebox("Slicer5:\n\nSave this SKP before using this tool...")
    return nil
  end
  @model.active_layer=nil
  @ss=@model.selection
  if @ss.empty?
    UI.messagebox("Slicer5:\n\nNO Selection !")
    return nil
  end#if
  if @ss[1]
    UI.messagebox("Slicer5:\n\nSelection MUST be ONE Group or Component-Instance !")
    return nil
  end#if
  if @ss[0].class != Sketchup::Group and @ss[0].class != Sketchup::ComponentInstance
    UI.messagebox("Slicer5:\n\nSelection is NOT a Group or Component-Instance !")
    return nil
  end#if
  @selected=@ss[0]
  ### check if 'manifold'
  if not @selected.manifold?
    UI.messagebox("Slicer5:\n\nThe Selected Object is NOT a 'Manifold' Solid !")
    return nil
  end#if
  ### get bb etc
  ###
  @trans=@selected.transformation
  @bb=@selected.bounds
  @max=@bb.max
  @min=@bb.min
  @maxZ=@max.z-@min.z
  @maxY=@max.y-@min.y
  @maxX=@max.x-@min.x
  @offZ=@maxZ/2.0
  @offY=@maxY/2.0
  @offX=@maxX/2.0
  @cent=@bb.center
  @maxZC=@cent.clone
  @maxZC.z=@max.z
  @maxYC=@cent.clone
  @maxYC.y=@max.y
  @maxXC=@cent.clone
  @maxXC.x=@max.x
  @diag=@bb.diagonal
### set up @@spacing1, @@thickness1 etc in def dialog()
### show VCB and status info
  Sketchup::set_status_text("", SB_VCB_LABEL)
  Sketchup::set_status_text("", SB_VCB_VALUE)
  @msg="Slicer5: "
  Sketchup.status_text=@msg+"Parameters..."
  ### get settings
  return nil unless self.dialog()
  ### run
  self.run()
  ###
end#initialize
##############

############
def dialog()
  if @model.options["UnitsOptions"]["LengthUnit"]<=1 ### it's '/"
    xspacing1=2.0.inch
    xthickness1=0.125.inch
    xinset_start1=0.inch
    xinset_end1=0.inch
	xtht=0.5.inch
  else ##mmm
    xspacing1=50.mm
    xthickness1=3.mm
    xinset_start1=0.mm
    xinset_end1=0.mm
	xtht=12.mm
  end
  ###
  axes=["X|Y|Z|XY|XZ|YZ"]
  refs=["Yes|No"]
  flatten=["Yes|No"]
  ###
  prompts=["Axis: ", "Spacing: ", "Thickness: ", "Inset at Start: ", "Inset at End: ", "Add References ? ", "Text Height: ", "Flatten ? "]
  ###
  title=@msg+"Parameters"
  ###
  @@axis="Z" unless @@axis
  @@spacing1=xspacing1 unless @@spacing1
  @@thickness1=xthickness1 unless @@thickness1
  @@inset_start1=xinset_start1 unless @@inset_start1
  @@inset_end1=xinset_end1 unless @@inset_end1
  @@ref="Yes" unless @@ref
  @@tht=xtht unless @tht
  @@flatten="Yes" unless @@flatten
  ###
  values=[@@axis, @@spacing1, @@thickness1, @@inset_start1, @@inset_end1, @@ref, @@tht, @@flatten] ###
  pops=[axes, "", "", "", "", refs, "", flatten] ###
  results=inputbox(prompts, values, pops, title)
  return nil unless results
  @@axis, @@spacing1, @@thickness1, @@inset_start1, @@inset_end1, @@ref, @@tht, @@flatten = results
  ###
if @@thickness1<0
   @@thickness1= -@@thickness1
   UI.messagebox("Slicer5:\n\nThickness CAN'T be < 0 !\nUsing +ve value = #{@@thickness1}")
end#if
if @@thickness1>@@spacing1 and @@spacing1>0
   if xthickness1<@@spacing1
      @@thickness1=xthickness1 
   else
      @@thickness1=@@spacing1
   end#if
   UI.messagebox("Slicer5:\n\nThickness CANNOT be > Spacing !\nThickness = Spacing = #{@@thickness1}")
end#if
if @@inset_start1<0
   @@inset_start1=xinset_start1
   UI.messagebox("Slicer5:\n\nInset at Start CAN'T be < 0 !\nUsing default = #{@@inset_start1}")
end#if
if @@inset_end1<0
   @@inset_end1=xinset_end1
   UI.messagebox("Slicer5:\n\nInset at End CAN'T be < 0 !\nUsing default = #{@@inset_end1}")
end#if
### sort out by spaces
if @@spacing1<=0
   @@spacing1=0.mm 
end#if
### sort out slice 2/slots etc
if @@axis=~/^[XY][YZ]$/ ### dual-axis
  return nil unless self.dialog1()
end
### spacing
if @@spacing1<=0 or (@vec2 and @@spacing2<=0)
  return nil unless self.dialog2()
end
###
  return true
  ###
end#dialog
########## #############################################################
def dialog1()
  ###
  Sketchup.status_text=@msg+"Parameters 2..."
  ###
    xspacing2=@@spacing1
    xthickness2=@@thickness1
    xinset_start2=@@inset_start1
    xinset_end2=@@inset_end1
	xtol=0.mm
	xbit=0.mm
  ###
  slots=["Yes|No"]
  ###
  prompts=["Spacing 2: ", "Thickness 2: ", "Inset at Start 2: ", "Inset at End 2: ", "Slot ? ","Tolerance: ", "Overcut: "]
  ###
  title=@msg+"Parameters 2"
  @@spacing2=xspacing2 unless @@spacing2
  @@thickness2=xthickness2 unless @@thickness2
  @@inset_start2=xinset_start2 unless @@inset_start2
  @@inset_end2=xinset_end2 unless @@inset_end2
  @@slot="Yes" unless @@slot
  @@slot="No" if @@thickness1==0
  @@tol=xtol unless @@tol
  @@bit=xbit unless @@bit
  values=[@@spacing2, @@thickness2, @@inset_start2, @@inset_end2, @@slot, @@tol, @@bit] ###
  popups=["", "", "", "", slots, "", ""] ###
  results=inputbox(prompts, values, popups, title)
  return nil unless results
  @@spacing2, @@thickness2, @@inset_start2, @@inset_end2, @@slot, @@tol, @@bit = results
  ###
  if @@spacing2<=0
    @@spacing2=0.mm
  end
  ###
  if @@thickness2<0
    @@thickness2= -@@thickness2
    UI.messagebox("Slicer5:\n\nThickness2 CAN'T be < 0 !\nUsing +ve value = #{@@thickness2}")
  end#if
  if @@thickness2>@@spacing2 and @@spacing2>0
    if xthickness12<@@spacing2
      @@thickness2=xthickness2 
    else
      @@thickness2=@@spacing2
    end#if
   UI.messagebox("Slicer5:\n\nThickness2 CANNOT be > Spacing2 !\nThickness2 = Spacing2 = #{@@thickness2}")
  end#if
  if @@inset_start2<0
     @@inset_start2=xinset_start2
     UI.messagebox("Slicer5:\n\nInset at Start 2 CAN'T be < 0 !\nUsing default = #{@@inset_start2}")
  end#if
  if @@inset_end2<0
    @@inset_end2=xinset_end2
    UI.messagebox("Slicer5:\n\nInset at End 2 CAN'T be < 0 !\nUsing default = #{@@inset_end2}")
  end#if
  if @@slot=="No"
    @@tol=xtol
	@@bit=xbit
  end
  ###
  return true
  ###
end#dialog1
########### ############################################################
def dialog2()
  ###
  Sketchup.status_text=@msg+"Spaces..."
  ###
  @@spaces1=1 if @@spaces1<1
  @@spaces2=1 if @@spaces2<1
  ###
  prompt2=nil
  case @@axis
    when "Z"
      prompt1="Spaces [Z "+@maxZ.to_l.to_s+"]: "
	  prompts=[prompt1]
	  values=[@@spaces1]
    when "Y"
      prompt1="Spaces [Y "+@maxY.to_l.to_s+"]: "
	  prompts=[prompt1]
	  values=[@@spaces1]
    when "X"
      prompt1="Spaces [X "+@maxX.to_l.to_s+"]: "
	  prompts=[prompt1]
	  values=[@@spaces1]
    when "XY"
      prompt1="Spaces [X "+@maxX.to_l.to_s+"]: "
      prompt2="Spaces [Y "+@maxY.to_l.to_s+"]: "
	  prompts=[prompt1,prompt2]
	  values=[@@spaces1, @@spaces2]
    when "XZ"
      prompt1="Spaces [X "+@maxX.to_l.to_s+"]: "
      prompt2="Spaces [Z "+@maxZ.to_l.to_s+"]: "
	  prompts=[prompt1,prompt2]
	  values=[@@spaces1, @@spaces2]
    when "YZ"
      prompt1="Spaces [Y "+@maxY.to_l.to_s+"]: "
      prompt2="Spaces [Z "+@maxZ.to_l.to_s+"]: "
	  prompts=[prompt1,prompt2]
	  values=[@@spaces1, @@spaces2]
	else ### ???
	  ### ???
  end
  title=@msg+"Spaces"
  results=inputbox(prompts,values,title)
  return nil unless results
  if prompt2
    @@spaces1, @@spaces2 = results
  else
    @@spaces1=results[0]
  end
  @@spaces1=1 if @@spaces1<1
  @@spaces2=1 if prompt2 and @@spaces2<1
  ###
  @spaces1=@@spaces1
  @spaces2=@@spaces2 if prompt2
  ###
  return true
  ###
end#dialog2
########### ############################################################

def layers_mats()
  ### make layers
  @layers=@model.layers
@layerXXXX=@layers.add("SLCR-XXXX")
@layerXXXX.visible=true
@layerYYYY=@layers.add("SLCR-YYYY")
@layerYYYY.visible=true
@layerZZZZ=@layers.add("SLCR-ZZZZ")
@layerZZZZ.visible=true
@layerFLAT=@layers.add("SLCR-FLAT")
@layerFLAT.visible=true
@layerTEXT=@layers.add("SLCR-TEXT")
@layerTEXT.visible=true
@layerHIDN=@layers.add("SLCR-HIDN")
@layerHIDN.visible=false
  ###
  @model.active_layer=nil
  ### make materials
  @mats=@model.materials
@matXXXX=@mats.add("SLCR-XXXX") unless @matXXXX=@mats["SLCR-XXXX"]
@matXXXX.color=[255,130,130]
@matXXXX.alpha=1.0
@matYYYY=@mats.add("SLCR-YYYY") unless @matYYYY=@mats["SLCR-YYYY"]
@matYYYY.color=[130,255,130]
@matYYYY.alpha=1.0
@matZZZZ=@mats.add("SLCR-ZZZZ") unless @matZZZZ=@mats["SLCR-ZZZZ"]
@matZZZZ.color=[130,130,255]
@matZZZZ.alpha=1.0
  ###
end#layers_mats
###
def run()
  ###
  @model.start_operation("Slicer5", true)
  ###
  GC.enable()
  GC.start() ### garbage collection...
  ###
  @ents=@model.active_entities
  ###
  self.layers_mats()
  ###
  @ss.clear
  ###
  @rnd=sprintf("%.4d", (rand*10000))
  ###
  @spac1=@@spacing1
  @thick1=@@thickness1
  @insS1=@@inset_start1
  @insE1=@@inset_end1
  @spac2=@@spacing2
  @thick2=@@thickness2
  @insS2=@@inset_start2
  @insE2=@@inset_end2
  ###
  if @@ref=="Yes"
    @ref=true
  else
    @ref=false
  end
  if @@flatten=="Yes"
    @flat=true
  else
    @flat=false
  end
  if @@slot=="Yes"
    @slot=true
  else ### No
    @slot=false
  end
	@tol=@@tol
	@bit=@@bit
  ###
  ### add central one then rep to start, repeat to end and insS1/insE1
  ### axes "X|Y|Z|XY|XZ|YZ"
  @vec1=nil
  @off1=nil
  @pref1=nil
  @vec2=nil
  @off2=nil
  @pref2=nil
  ###
  case @@axis
    when "Z"
	  @vec1=Geom::Vector3d.new(0,0,@spac1)
	  @vec1=Geom::Vector3d.new(0,0,1) if @spac1==0
	  @off1=@offZ
	  @pref1="Z"
	  @slot=false
    when "Y"
	  @vec1=Geom::Vector3d.new(0,@spac1,0)
	  @vec1=Geom::Vector3d.new(0,1,0) if @spac1==0
	  @off1=@offY
	  @pref1="Y"
	  @slot=false
    when "X"
	  @vec1=Geom::Vector3d.new(@spac1,0,0)
	  @vec1=Geom::Vector3d.new(1,0,0) if @spac1==0
	  @off1=@offX
	  @pref1="X"
	  @slot=false
	when "XZ"
	  @vec1=Geom::Vector3d.new(@spac1,0,0)
	  @vec1=Geom::Vector3d.new(1,0,0) if @spac1==0
	  @off1=@offX
	  @pref1="X"
	  @vec2=Geom::Vector3d.new(0,0,@spac2)
	  @vec2=Geom::Vector3d.new(0,0,1) if @spac2==0
	  @off2=@offZ
	  @pref2="Z"
    when "XY"
	  @vec1=Geom::Vector3d.new(@spac1,0,0)
	  @vec1=Geom::Vector3d.new(1,0,0) if @spac1==0
	  @off1=@offX
	  @pref1="X"
	  @vec2=Geom::Vector3d.new(0,@spac2,0)
	  @vec2=Geom::Vector3d.new(0,1,0) if @spac2==0
	  @off2=@offY
	  @pref2="Y"
    when "YZ"
	  @vec1=Geom::Vector3d.new(0,@spac1,0)
	  @vec1=Geom::Vector3d.new(0,1,0) if @spac1==0
	  @off1=@offY
	  @pref1="Y"
	  @vec2=Geom::Vector3d.new(0,0,@spac2)
	  @vec2=Geom::Vector3d.new(0,0,1) if @spac2==0
	  @off2=@offZ
	  @pref2="Z"
  end
  @slicesS=[]
  @slicesE=[]
  @slicesS2=[]
  @slicesS2=[]
  @sliceC=[]
  @slices=[]
  ###
  Sketchup.status_text=@msg+"Slicing..."
  self.slice(@vec1)
  @slices1=@slices.dup
  ###
  if @vec2
    @slicesS=[]
    @slicesE=[]
    @slicesS2=[]
    @slicesS2=[]
    @sliceC=[]
    @slices=[]
	###
	@off1=@off2
	###
    Sketchup.status_text=@msg+"Slicing 2..."
    self.slice(@vec2)
	###
	@slices2=@slices.dup
    @group2.name=@group2.name.gsub(/^SLICES/,"SLICES2")
	###
  end
  ###
  @slices=@slices1.dup
  Sketchup.status_text=@msg+"Slotting..."
  self.slot() if @slot ### gp1
  ###
  if @vec2 and @slot ### gp2
    @slices=@slices2.dup
    Sketchup.status_text=@msg+"Slotting 2..."
    self.slot2()
  end
  ###
  @slices=@slices1.dup
  if @thick1>0
    Sketchup.status_text=@msg+"Fattening..."
    self.fatten(@vec1) ### +tol if tol>0
  end
  ###
  Sketchup.status_text=@msg+"Originating..."
  self.originate(@vec1)
  Sketchup.status_text=@msg+"Referencing..."
  self.ref(@vec1,@pref1)
  Sketchup.status_text=@msg+"Making Components..."
  self.componentify()
  Sketchup.status_text=@msg+"Flattening..."
  self.flatten(@vec1) if @flat
  ###
  if @vec2
    @slices=@slices2.dup
	@off1=@off2
	###
    if @thick2>0
	  Sketchup.status_text=@msg+"Fattening 2..."
	  self.fatten(@vec2)
	end
	Sketchup.status_text=@msg+"Originating 2..."
	self.originate(@vec2)
	Sketchup.status_text=@msg+"Referencing 2..."
    self.ref(@vec2,@pref2)
	Sketchup.status_text=@msg+"Making Components 2..."
	self.componentify()
	 Sketchup.status_text=@msg+"Flattening..."
	self.flatten(@vec2) if @flat
  end
  ###
  @selected.layer=@layerHIDN
  ### export ?
  @model.active_view.refresh
  if @flat
    Sketchup.status_text=@msg+"Export ?"
	@dxf=false
    TIG::Slicer.export(@rnd, @flat1, @flat2)
  end
  ###
  Sketchup.status_text=@msg+"Done."
  ###
  @model.commit_operation unless @dxf
  ###
end#run

def slice(vec=Z_AXIS)
  gp=@ents.add_group
  gents=gp.entities
  gp.name="SLICES-"+@rnd
  if vec==@vec2
    @group2=gp
	if @spac2==0
	  vec.length=((@off1*2.0)-@insS2-@insE2-@thick2)/@spaces2
	  spaces=@spaces2
	  spaced=true
	else
	  spaced=false
	end
  else
    @group1=gp
	if @spac1==0
	  vec.length=((@off1*2.0)-@insS1-@insE1-@thick1)/@spaces1
	  spaced=true
	  spaces=@spaces1
	else
	  spaced=false
	end
  end
  ###
 if spaced #############################################################
  ###
  if vec.parallel?(Z_AXIS)
    gp.layer=@layerZZZZ
	gp.material=@matZZZZ
    pt=@maxZC.clone
	pt.z=pt.z-@insS1-(@thick1/2.0) if vec.parallel?(@vec1)
	pt.z=pt.z-@insS2-(@thick2/2.0) if @vec2 and vec.parallel?(@vec2)
	vec.reverse!
  elsif vec.parallel?(Y_AXIS)
    gp.layer=@layerYYYY
	gp.material=@matYYYY
    pt=@maxYC.clone
	pt.y=pt.y-@insS1-(@thick1/2.0) if vec.parallel?(@vec1)
	pt.y=pt.y-@insS2-(@thick2/2.0) if @vec2 and vec.parallel?(@vec2)
	vec.reverse!
  else ### X
    gp.layer=@layerXXXX
	gp.material=@matXXXX
    pt=@maxXC.clone
	pt.x=pt.x-@insS1-(@thick1/2.0) if vec.parallel?(@vec1)
    pt.x=pt.x-@insS2-(@thick2/2.0) if @vec2 and vec.parallel?(@vec2)
	vec.reverse!
  end
  (spaces+1).times{
    tgp=gents.add_group()
    tents=tgp.entities
    eds=tents.add_circle(pt, vec, @diag, 12)
    fac=tents.add_face(eds)
    fac.reverse! if fac.normal!=vec.normalize
    tr0=tgp.transformation
    tr0=Geom::Transformation.new()
    tents.intersect_with(true,tr0,tents,tr0,true,@selected)
    tents.erase_entities(eds)
    self.tidy(tents,tr0,tgp)
    @slices << tgp
	pt.offset!(vec)
  }
  ###
 else ### spacing ######################################################
  ###
  tgp=gents.add_group()
  tents=tgp.entities
  eds=tents.add_circle(@cent, vec, @diag, 12)
  fac=tents.add_face(eds)
  fac.reverse! if fac.normal!=vec.normalize
  tr0=tgp.transformation
  tr0=Geom::Transformation.new()
  tents.intersect_with(true,tr0,tents,tr0,true,@selected)
  tents.erase_entities(eds)
  ### clean up
  self.tidy(tents,tr0,tgp)
  ###
  @sliceC << tgp ### make central one 1st
  ###
 if vec==@vec2 ##### gp2
  ###
  if vec.parallel?(Z_AXIS)
    gp.layer=@layerZZZZ
	gp.material=@matZZZZ
    repS=((@offZ-@insS2-(@thick2/2.0))/@spac2).to_i
    repE=((@offZ-@insE2-(@thick2/2.0))/@spac2).to_i
  elsif vec.parallel?(Y_AXIS)
    gp.layer=@layerYYYY
	gp.material=@matYYYY
    repS=((@offY-@insS2-(@thick2/2.0))/@spac2).to_i
    repE=((@offY-@insE2-(@thick2/2.0))/@spac2).to_i
  else ### X
    gp.layer=@layerXXXX
	gp.material=@matXXXX
    repS=((@offX-@insS2-(@thick2/2.0))/@spac2).to_i
    repE=((@offX-@insE2-(@thick2/2.0))/@spac2).to_i
  end
  insS=@insS2
  insE=@insE2
  thick=@thick2
  ###
 else ########## gp1
  ###
  if vec.parallel?(Z_AXIS)
    gp.layer=@layerZZZZ
	gp.material=@matZZZZ
    repS=((@offZ-@insS1-(@thick1/2.0))/@spac1).to_i
    repE=((@offZ-@insE1-(@thick1/2.0))/@spac1).to_i
  elsif vec.parallel?(Y_AXIS)
    gp.layer=@layerYYYY
	gp.material=@matYYYY
    repS=((@offY-@insS1-(@thick1/2.0))/@spac1).to_i
    repE=((@offY-@insE1-(@thick1/2.0))/@spac1).to_i
  else ### X
    gp.layer=@layerXXXX
	gp.material=@matXXXX
    repS=((@offX-@insS1-(@thick1/2.0))/@spac1).to_i
    repE=((@offX-@insE1-(@thick1/2.0))/@spac1).to_i
  end
  insS=@insS1
  insE=@insE1
  thick=@thick1
  ###
 end #####
  ### to Start
  pt=@cent.clone
  (repS).times{|i|
	pt.offset!(vec)
    tgp=gents.add_group()
    tents=tgp.entities
    eds=tents.add_circle(pt, vec, @diag, 12)
    fac=tents.add_face(eds)
    fac.reverse! if fac.normal!=vec.normalize
    tr0=tgp.transformation
	tr0=Geom::Transformation.new()
    tents.intersect_with(true,tr0,tents,tr0,true,@selected)
    tents.erase_entities(eds)
	### clean up
    self.tidy(tents,tr0,tgp)
    ###
	@slicesS << tgp
  }#times
  ###
  tvec=vec.clone
  tvec.length=@off1-insS-(thick/2.0)
  pt=@cent.offset(tvec)
    tgp=gents.add_group()
    tents=tgp.entities
    eds=tents.add_circle(pt, vec, @diag, 12)
    fac=tents.add_face(eds)
    fac.reverse! if fac.normal!=vec.normalize
    tr0=tgp.transformation
	tr0=Geom::Transformation.new()
    tents.intersect_with(true,tr0,tents,tr0,true,@selected)
    tents.erase_entities(eds)
	### clean up
    self.tidy(tents,tr0,tgp)
  @slicesS << tgp
  ### to End
  pt=@cent.clone
  (repE).times{|i|
	pt.offset!(vec.reverse)
    tgp=gents.add_group()
    tents=tgp.entities
    eds=tents.add_circle(pt, vec, @diag, 12)
    fac=tents.add_face(eds)
    fac.reverse! if fac.normal!=vec.normalize
    tr0=tgp.transformation
	tr0=Geom::Transformation.new()
    tents.intersect_with(true,tr0,tents,tr0,true,@selected)
    tents.erase_entities(eds)
	### clean up
    self.tidy(tents,tr0,tgp)
    ###
	@slicesE << tgp
  }#times
  ###
  tvec=vec.clone.reverse
  tvec.length=@off1-insS-(thick/2.0)
  pt=@cent.offset(tvec)
    tgp=gents.add_group()
    tents=tgp.entities
    eds=tents.add_circle(pt, vec, @diag, 12)
    fac=tents.add_face(eds)
    fac.reverse! if fac.normal!=vec.normalize
    tr0=tgp.transformation
	tr0=Geom::Transformation.new()
    tents.intersect_with(true,tr0,tents,tr0,true,@selected)
    tents.erase_entities(eds)
	### clean up
    self.tidy(tents,tr0,tgp)
  @slicesE << tgp
  ### assemble groups in order and number etc...
  @slices = @slicesS.reverse + @sliceC + @slicesE
  ###
 end#if spaced #########################################################
  ###
end#slice


def tidy(tents=[],tr0=Geom::Transformation.new(),tgp=nil)
  4.times{
    tents.to_a.each{|e|e.find_faces if e.class==Sketchup::Edge}
    tents.intersect_with(true,tr0,tents,tr0,true,tents.to_a)
  }#times
  faces=[]
  tents.each{|e|faces << e if e.class==Sketchup::Face}
  (faces.length-1).times{
   for face in faces
    if face.valid?
      for edgeuse in face.outer_loop.edgeuses
         if not edgeuse.partners[0] ### outermost face
            faces=faces-[face]
            loops=face.loops
            for loop in loops
               for fac in faces
                  if fac.valid? and (fac.outer_loop.edges-loop.edges)==[]
                     faces=faces-[fac]
                     fac.erase! if fac.valid? ### fac abutts kept face so it must be erased...
                  end #if fac
               end #for fac
            end #for loop
         end #if outermost
      end #for edgeuse
    end #if valid
   end #for face
  }#times
end#tidy


def fatten(vec=Z_AXIS)
  if vec==@vec2
    thick=(@thick2/2.0)
  else
    thick=(@thick1/2.0)
  end
  @slices.each{|gp|
	gents=gp.entities
    faces=[]
	gents.each{|e|faces << e if e.class==Sketchup::Face}
	faces.each{|face|face.pushpull(thick)}
	faces=[]
	gents.each{|e|faces << e if e.class==Sketchup::Face and e.normal==vec.reverse.normalize}
	faces.each{|face|face.pushpull(thick)}
  }
end#fatten


def originate(vec=Z_AXIS)
  @slices.each{|gp|
	gents=gp.entities
    bb=gp.bounds
	if vec.parallel?(Z_AXIS)
	  ### relocate
      pt=bb.min
	  ve=pt.vector_to(ORIGIN)
	  tr=Geom::Transformation.translation(ve)
	  gents.transform_entities(tr,gents.to_a)
	  gp.transform!(tr.inverse)
	  gp.entities.parent.invalidate_bounds
    elsif vec.parallel?(Y_AXIS)
	  ### relocate
      pt=bb.min
	  pt.y=pt.y+@thick1
	  ve=pt.vector_to(ORIGIN)
	  tr=Geom::Transformation.translation(ve)
	  gents.transform_entities(tr,gents.to_a)
	  gp.transform!(tr.inverse)
	  gp.entities.parent.invalidate_bounds
	  ### rotate
	  gpt=gp.transformation
	  opt=gpt.origin
	  tr=Geom::Transformation.rotation(ORIGIN, X_AXIS, -90.degrees)
	  gents.transform_entities(tr,gents.to_a)
	  tr=Geom::Transformation.rotation(opt, X_AXIS, 90.degrees)
	  gp.transform!(tr)
	  gp.entities.parent.invalidate_bounds
    else ### X
	  ### relocate
      pt=bb.min
	  ve=pt.vector_to(ORIGIN)
	  tr=Geom::Transformation.translation(ve)
	  gents.transform_entities(tr,gents.to_a)
	  gp.transform!(tr.inverse)
	  gp.entities.parent.invalidate_bounds
	  ### rotate
	  gpt=gp.transformation
	  opt=gpt.origin
	  tr=Geom::Transformation.rotation(ORIGIN, Y_AXIS, -90.degrees)
	  gents.transform_entities(tr,gents.to_a)
	  tr=Geom::Transformation.rotation(opt, Y_AXIS, 90.degrees)
	  gp.transform!(tr)
	  gp.entities.parent.invalidate_bounds
    end
  }
end

def ref(vec=Z_AXIS, pref="?")
  ### rename groups... then perhaps add ref code if @ref
  ###
  num=pref+"001"
  ###
  @slices.each{|gp|
    gp.name=num
	num.next!
	if @ref
      txt=gp.name
	  txgp=gp.entities.add_group()
	  txgp.name="TEXT"
	  txgp.layer=@layerTEXT
	  txgp.entities.add_3d_text(txt,TextAlignLeft,'Txt',false,false,@@tht,0,0,false,0)
	  tt=[@@tht, @@tht, @thick1]
	  tm=txgp.bounds.min
	  ve=tm.vector_to(tt)
	  tr=Geom::Transformation.translation(ve)
	  txgp.transform!(tr)
	  if vec.parallel?(X_AXIS)
	    tr=Geom::Transformation.rotation(ORIGIN,Z_AXIS,90.degrees)
		txgp.transform!(tr)
	  else
	    ### do nothing because it's already Y or Z
	  end
    end
  }
end#ref

def componentify()
  @defs=[]
  @slices.each{|gp|
    name=gp.name
    ins=gp.to_component
	defn=ins.definition
	defn.name=@rnd+"-"+name
	@defs << defn
  }
end#componentify

def flatten(vec=Z_AXIS)
  gp=@ents.add_group()
  gp.name="FLAT-"+@rnd
  gp.layer=@layerFLAT
  @flat1=gp if vec!=@vec2
  nudge=false
  if vec.parallel?(Z_AXIS)
	gp.material=@matZZZZ
  elsif vec.parallel?(Y_AXIS)
	gp.material=@matYYYY
  else ### X
	gp.material=@matXXXX
	nudge=true
  end
  gents=gp.entities
  pt=@selected.bounds.max
  pt.z=0
  pt.x=pt.x+@selected.bounds.depth if vec.parallel?(X_AXIS)
  if vec==@vec2 ### put seconf flat set UP
    pt.y=pt.y+@thick1+@y2-@selected.bounds.max.y
	gp.name="FLAT2-"+@rnd
	@flat2=gp
  end
  pt.offset!([@thick1,@thick1,0])
  ### arranger
  num=@defs.length
  numx=Math::sqrt(num).round
  numx=1 if numx<1
  ctrx=0
  px=pt.x
  py=pt.y
  ###
  @defs.each{|defn|
    tr=Geom::Transformation.new(pt)
    ins=gents.add_instance(defn, tr)
	pt.x=pt.x+@thick1+ins.bounds.width
	ctrx+=1
	if ctrx>=numx
	  gp.entities.parent.invalidate_bounds
	  pt.x=px
	  pt.y=py+@thick1+gp.bounds.height
	  ctrx=0
	end
  }
  gp.entities.parent.invalidate_bounds
  @y2=@thick1+gp.bounds.max.y
end#flatten

##############################
def delete_unwanted_edges(ents)
  ###
  edges=[]
  ents.each{|e|
	edges << e if e.valid? and e.kind_of?(Sketchup::Edge)
  }
  togos=[]
  edges.each{|e|
    next unless e.valid?
    if e.faces.length==0
      togos << e
	elsif e.faces.length==2 and e.faces[0].normal.dot(e.faces[1].normal) > 0.99999999
	  togos << e
    end
  }
  ents.erase_entities(togos)
  ###
end
############################

def slot()
  ### slot gp1 with gp2
  @group1.entities.to_a.each{|gp|
    next unless gp.class==Sketchup::Group
    gents=gp.entities
	tr0=Geom::Transformation.new()
	trg=gp.transformation
	eds=gents.intersect_with(true, trg, gents, tr0, true, @group2)
    ### replace with half length line
	neds=[]
	eds.each{|e|
	  next unless e.valid?
	  ve=e.line[1]
	  mpt=e.start.position.offset(ve, e.length/2)
	  neds << mpt
	  e.erase! if e.valid?
	}
	neds.each{|mpt|
	  ###
	  if @vec1.parallel?(X_AXIS)
	    ###
	    if @vec2.parallel?(Y_AXIS) ### X & Y
	      p0=mpt.offset([0, -((@thick2/2.0)+(@tol/2.0)), 0])
		  p1=p0.offset([0, @thick2+@tol, 0])
		  ### move p0/p1 by bit/2
		  if @bit>0
		    vp=[0, 0, -@bit/2.0]
			p0.offset!(vp)
			p1.offset!(vp)
		  end
		  ga=gents.to_a
		  ln=gents.add_line(p0, p1) ### slot end
		  ray=@model.raytest([p0, [0, 0, 1]])
		  if ray
		    p00=ray[0]
		  else
		    p00=p0.offset([0, 0, @diag])
		  end
		  e00=gents.add_line(p0, p00)
		  ray=@model.raytest([p1, [0, 0, 1]])
		  if ray
		    p11=ray[0]
		  else
		    p11=p1.offset([0, 0, @diag])
		  end
		  e11=gents.add_line(p1, p11)
		  gents.intersect_with(true, tr0, gents, tr0, true, ga)
		  ln.find_faces
		  if e00 and e00.valid? and e00.faces[1]
			if e00.faces[0].area<e00.faces[1].area
			  e00.faces[0].erase!
			else
			  e00.faces[1].erase!
			end
		  end
		  if e11 and e11.valid? and e11.faces[1]
			if e11.faces[0].area<e11.faces[1].area
			  e11.faces[0].erase!
			else
			  e11.faces[1].erase!
			end
		  end
		  ###
		else ### X & Z
		  ###
	      p0=mpt.offset([0 ,0, -((@thick2/2.0)+(@tol/2.0))])
		  p1=p0.offset([0, 0, @thick2+@tol])
		  ### move p0/p1 by bit/2
		  if @bit>0
		    vp=[0, -@bit/2.0, 0]
			p0.offset!(vp)
			p1.offset!(vp)
		  end
		  ga=gents.to_a
		  ln=gents.add_line(p0, p1) ### slot end
		  ray=@model.raytest([p0, [0, 1, 0]])
		  if ray
		    p00=ray[0]
		  else
		    p00=p0.offset([0, @diag, 0])
		  end
		  e00=gents.add_line(p0, p00)
		  ray=@model.raytest([p1, [0, 1, 0]])
		  if ray
		    p11=ray[0]
		  else
		    p11=p1.offset([0, @diag, 0])
		  end
		  e11=gents.add_line(p1, p11)
		  gents.intersect_with(true, tr0, gents, tr0, true, ga)
		  ln.find_faces
		  if e00 and e00.valid? and e00.faces[1]
			if e00.faces[0].area<e00.faces[1].area
			  e00.faces[0].erase!
			else
			  e00.faces[1].erase!
			end
		  end
		  if e11 and e11.valid? and e11.faces[1]
			if e11.faces[0].area<e11.faces[1].area
			  e11.faces[0].erase!
			else
			  e11.faces[1].erase!
			end
		  end
		  ###
		end
	  ###
	  else ### parallel to Y_AXIS >>> Y & Z only
	  ###
	    ###
	      p0=mpt.offset([0, 0, -((@thick2/2.0)+(@tol/2.0))])
		  p1=p0.offset([0, 0, @thick2+@tol])
		  ### move p0/p1 by bit/2
		  if @bit>0
		    vp=[-@bit/2.0, 0, 0]
			p0.offset!(vp)
			p1.offset!(vp)
		  end
		  ga=gents.to_a
		  ln=gents.add_line(p0, p1) ### slot end
		  ray=@model.raytest([p0, [1, 0, 0]])
		  if ray
		    p00=ray[0]
		  else
		    p00=p0.offset([@diag, 0, 0])
		  end
		  e00=gents.add_line(p0, p00)
		  ray=@model.raytest([p1, [1, 0, 0]])
		  if ray
		    p11=ray[0]
		  else
		    p11=p1.offset([@diag, 0, 0])
		  end
		  e11=gents.add_line(p1, p11)
		  gents.intersect_with(true, tr0, gents, tr0, true, ga)
		  ln.find_faces
		  if e00 and e00.valid? and e00.faces[1]
			if e00.faces[0].area<e00.faces[1].area
			  e00.faces[0].erase!
			else
			  e00.faces[1].erase!
			end
		  end
		  if e11 and e11.valid? and e11.faces[1]
			if e11.faces[0].area<e11.faces[1].area
			  e11.faces[0].erase!
			else
			  e11.faces[1].erase!
			end
		  end
		###
	  end
	  ###
	}
	###
    self.delete_unwanted_edges(gents)
	###
  }
 ###
end#slot

def slot2()
  ### slot gp2 with gp1
  @group2.entities.to_a.each{|gp|
    next unless gp.valid? and gp.class==Sketchup::Group
    gents=gp.entities
	tr0=Geom::Transformation.new()
	trg=gp.transformation
	eds=gents.intersect_with(true, trg, gents, tr0, true, @group1)
    ### replace with half length line
	neds=[]
	eds.each{|e|
	  next unless e.valid?
	  if @vec2.parallel?(Y_AXIS) ### || X >>> z
	    if e.start.position.z > e.end.position.z
	      mpt=e.start.position
		else
		  mpt=e.end.position
		end
	  else ### || Z
	    if @vec1.parallel?(X_AXIS) ### || Z >>> y
	      if e.start.position.y > e.end.position.y
	        mpt=e.start.position
		  else
		    mpt=e.end.position
		  end
		else ### vec1 || Y >>> x
		  if e.start.position.x > e.end.position.x
	        mpt=e.start.position
		  else
		    mpt=e.end.position
		  end
		end
	  end
	  neds << mpt
	  e.erase! if e.valid?
	}
	neds.each{|mpt|
	  ###
	  if @vec2.parallel?(Y_AXIS) ### with X
	    ###
	      p0=mpt.offset([-((@thick2/2.0)+(@tol/2.0)), 0, 0])
		  p1=p0.offset([@thick2+@tol, 0, 0])
		  ### move p0/p1 by bit/2
		  if @bit>0
		    vp=[0, 0, @bit/2.0]
			p0.offset!(vp)
			p1.offset!(vp)
		  end
		  ga=gents.to_a
		  ln=gents.add_line(p0, p1) ### slot end
		  ray=@model.raytest([p0, [0, 0, -1]])
		  if ray
		    p00=ray[0]
		  else
		    p00=p0.offset([0, 0, -@diag])
		  end
		  e00=gents.add_line(p0, p00)
		  ray=@model.raytest([p1, [0, 0, -1]])
		  if ray
		    p11=ray[0]
		  else
		    p11=p1.offset([0, 0, -@diag])
		  end
		  e11=gents.add_line(p1, p11)
		  gents.intersect_with(true, tr0, gents, tr0, true, ga)
		  ln.find_faces
		  if e00 and e00.valid? and e00.faces[1]
			if e00.faces[0].area<e00.faces[1].area
			  e00.faces[0].erase!
			else
			  e00.faces[1].erase!
			end
		  end
		  if e11 and e11.valid? and e11.faces[1]
			if e11.faces[0].area<e11.faces[1].area
			  e11.faces[0].erase!
			else
			  e11.faces[1].erase!
			end
		  end
		###
	  else ### Z & X OR Y
	    ###
		if @vec1.parallel?(X_AXIS) ### X & Z
		  ###
	      p0=mpt.offset([-((@thick2/2.0)+(@tol/2.0)), 0, 0])
		  p1=p0.offset([@thick2+@tol, 0, 0])
		  ### move p0/p1 by bit/2
		  if @bit>0
		    vp=[0, @bit/2.0, 0]
			p0.offset!(vp)
			p1.offset!(vp)
		  end
		  ga=gents.to_a
		  ln=gents.add_line(p0, p1) ### slot end
		  ray=@model.raytest([p0, [0, -1, 0]])
		  if ray
		    p00=ray[0]
		  else
		    p00=p0.offset([0, -@diag, 0])
		  end
		  e00=gents.add_line(p0, p00)
		  ray=@model.raytest([p1, [0, -1, 0]])
		  if ray
		    p11=ray[0]
		  else
		    p11=p1.offset([0, -@diag, 0])
		  end
		  e11=gents.add_line(p1, p11)
		  gents.intersect_with(true, tr0, gents, tr0, true, ga)
		  ln.find_faces
		  if e00 and e00.valid? and e00.faces[1]
			if e00.faces[0].area<e00.faces[1].area
			  e00.faces[0].erase!
			else
			  e00.faces[1].erase!
			end
		  end
		  if e11 and e11.valid? and e11.faces[1]
			if e11.faces[0].area<e11.faces[1].area
			  e11.faces[0].erase!
			else
			  e11.faces[1].erase!
			end
		  end
		###
		else ### Y & Z
		###
	      p0=mpt.offset([0, -((@thick2/2.0)+(@tol/2.0)), 0])
		  p1=p0.offset([0, @thick2+@tol, 0])
		  ### move p0/p1 by bit/2
		  if @bit>0
		    vp=[@bit/2.0, 0, 0]
			p0.offset!(vp)
			p1.offset!(vp)
		  end
		  ga=gents.to_a
		  ln=gents.add_line(p0, p1) ### slot end
		  ray=@model.raytest([p0, [-1, 0, 0]])
		  if ray
		    p00=ray[0]
		  else
		    p00=p0.offset([-@diag, 0, 0])
		  end
		  e00=gents.add_line(p0, p00)
		  ray=@model.raytest([p1, [-1, 0, 0]])
		  if ray
		    p11=ray[0]
		  else
		    p11=p1.offset([-@diag, 0, 0])
		  end
		  e11=gents.add_line(p1, p11)
		  gents.intersect_with(true, tr0, gents, tr0, true, ga)
		  ln.find_faces
		  if e00 and e00.valid? and e00.faces[1]
			if e00.faces[0].area<e00.faces[1].area
			  e00.faces[0].erase!
			else
			  e00.faces[1].erase!
			end
		  end
		  if e11 and e11.valid? and e11.faces[1]
			if e11.faces[0].area<e11.faces[1].area
			  e11.faces[0].erase!
			else
			  e11.faces[1].erase!
			end
		  end
		  ###
		end
	  end
	}
	###
	self.delete_unwanted_edges(gents)
	###
  }
  ###
end#slot2

def self.export(rnd=nil, flat1=nil, flat2=nil)
  @msg="Slicer5: "
  @model=Sketchup.active_model
  @ents=@model.entities
  @defns=@model.definitions
  @ss=@model.selection
  @rnd=rnd
  @flat1=flat1
  @flat2=flat2
  ### sort out rnd/flat1/flat2
  @sep=false
  @ex=""
  unless @rnd ### it's done separately
    @sep=true
  	@ex=" Exporter"
    @msg="Slicer5"+@ex+": "
	### get flat1/2 and rnd from ss
	if @ss.empty?
      UI.messagebox("Slicer5 Exporter:\n\nNO Selection !")
      return nil
    end#if
    if @ss[1]
      UI.messagebox("Slicer5 Exporter:\n\nSelection MUST be ONE 'FLAT-NNNN' Group !")
      return nil
    end#if
    if @ss[0].class != Sketchup::Group or not @ss[0].name =~ /^FLAT[-][0-9][0-9][0-9][0-9]/
      UI.messagebox("Slicer5 Exporter:\n\nSelection is NOT a 'FLAT-NNNN' Group !")
      return nil
    end#if
    @flat1=@ss[0]
	@ss.clear
	f1name=@flat1.name
	###
	@rnd=f1name.split("-")[1]
	###
	f2name=f1name.gsub(/[-]/,"2-")
	@flat2=nil
	@defns.each{|d|
	  next unless d.group?
	  if d.instances[0].name==f2name
	    @flat2=d.instances[0]
		break
	  end
	}
  end
  UI.beep
  ans=UI.messagebox("Slicer5"+@ex+":\n\nDo you want to Export these Slices as PNG Image Files ?",MB_YESNO,"")
  if ans==6
	self.png()
  end
  if Sketchup.is_pro?
	UI.beep ### DXF
	ans=UI.messagebox("Slicer5"+@ex+":\n\nDo you want to Export these Slices as DXF Files ?",MB_YESNO,"")
	if ans==6
	  self.dxf()
	end
  end
  ###
end#export

def self.png()
  @ss.clear
  imgfolder=File.join(File.dirname(@model.path), @model.title+"-Slicer5-PNGs-"+@rnd).tr("\\","/")
  Dir.mkdir(imgfolder) if not File.exist?(imgfolder)
  cam=@model.active_view.camera
  ppx=cam.perspective? rescue ppx=false
  tcam=Sketchup::Camera.new(cam.eye, cam.target, cam.up, ppx)
  cam.perspective=false
  ### hide all
  all=@model.entities.to_a+@model.active_entities.to_a-[@flat1]
  all.uniq!
  all.dup.each{|e|all.delete(e)if e.hidden?}
  all.each{|e|e.hidden=true}
  ### do flat1
  @model.active_view.refresh
  ints=@flat1.entities.to_a
  ### find max size
  mwidth=0
  mheight=0
  ints.each{|ins|
    ins.definition.invalidate_bounds
    bb=ins.bounds
    mwidth=bb.width if bb.width > mwidth
    mheight=bb.height if bb.height > mheight
  }
  vpw=@model.active_view.vpwidth
  vph=@model.active_view.vpheight
  count='0'
  tot=ints.length.to_s
  ints.each{|ins|
    ints.each{|e|e.hidden=false}
    imgname=ins.definition.name+".png"
    imgpath=File.join(imgfolder,imgname).tr("\\","/")
    ints.each{|e|e.hidden=true if not e==ins}
    bb=ins.bounds
    ctr=bb.center
    ctr.transform!(@flat1.transformation.inverse)
    ctr.transform!(ins.transformation.inverse)
    pmin=ctr.offset([-mwidth/2.0, -mheight/2.0, 0])
    pmax=ctr.offset([mwidth/2.0, mheight/2.0, 0])
    ctrp=ins.definition.entities.add_cpoint(ctr)
    cmin=ins.definition.entities.add_cpoint(pmin)
    cmax=ins.definition.entities.add_cpoint(pmax)
    @ss.clear
    @ss.add(ins)
    ins.definition.invalidate_bounds
    ctr.transform!(@flat1.transformation)
    ctr.transform!(ins.transformation)
    cam.set([ctr.x, ctr.y, ctr.z+1], ctr, [0,1,0])
    Sketchup.send_action("viewZoomToSelection:")
    ins.definition.entities.erase_entities([ctrp,cmin,cmax])
    if count=='0'
      UI.messagebox(@msg+"\n\n"+tot+" "+@flat1.name+" PNG Image Files into\n\n"+imgfolder)
    end#if
    count.next!
    @ss.clear
    @model.active_view.refresh
    keys={
        :filename => imgpath,
        :width => vpw,
        :height => vph,
        :antialias => false,
        :compression => 0.9
    }
    @model.active_view.write_image(keys)
  }
  ints.each{|e|e.hidden=false}
  @ss.clear
  ### unhide
  all.each{|e|e.hidden=false}
  ### do flat2
 if @flat2
   ### hide all
  all=@model.entities.to_a+@model.active_entities.to_a-[@flat2]
  all.uniq!
  all.dup.each{|e|all.delete(e)if e.hidden?}
  all.each{|e|e.hidden=true}
  ###
  @model.active_view.refresh
  ints=@flat2.entities.to_a
  ### find max size
  mwidth=0
  mheight=0
  ints.each{|ins|
    ins.definition.invalidate_bounds
    bb=ins.bounds
    mwidth=bb.width if bb.width > mwidth
    mheight=bb.height if bb.height > mheight
  }
  vpw=@model.active_view.vpwidth
  vph=@model.active_view.vpheight
  count='0'
  tot=ints.length.to_s
  ints.each{|ins|
    ints.each{|e|e.hidden=false}
    imgname=ins.definition.name+".png"
    imgpath=File.join(imgfolder,imgname).tr("\\","/")
    ints.each{|e|e.hidden=true if not e==ins}
    bb=ins.bounds
    ctr=bb.center
    ctr.transform!(@flat2.transformation.inverse)
    ctr.transform!(ins.transformation.inverse)
    pmin=ctr.offset([-mwidth/2.0, -mheight/2.0, 0])
    pmax=ctr.offset([mwidth/2.0, mheight/2.0, 0])
    ctrp=ins.definition.entities.add_cpoint(ctr)
    cmin=ins.definition.entities.add_cpoint(pmin)
    cmax=ins.definition.entities.add_cpoint(pmax)
    @ss.clear
    @ss.add(ins)
    ins.definition.invalidate_bounds
    ctr.transform!(@flat2.transformation)
    ctr.transform!(ins.transformation)
    cam.set([ctr.x, ctr.y, ctr.z+1], ctr, [0,1,0])
    Sketchup.send_action("viewZoomToSelection:")
    ins.definition.entities.erase_entities([ctrp,cmin,cmax])
    if count=='0'
      UI.messagebox(@msg+"\n\n"+tot+" "+@flat2.name+" PNG Image Files into\n\n"+imgfolder)
    end#if
    count.next!
    @ss.clear
    @model.active_view.refresh
    keys={
        :filename => imgpath,
        :width => vpw,
        :height => vph,
        :antialias => false,
        :compression => 0.9
    }
    @model.active_view.write_image(keys)
  }
  ints.each{|e|e.hidden=false}
  @ss.clear
  ### unhide
  all.each{|e|e.hidden=false}
 end#if
  ###
  tcam.perspective=ppx
  @model.active_view.camera=tcam
  Sketchup.send_action("viewZoomToSelection:")
  @model.active_view.refresh
  #UI.openURL('file:///'+imgfolder)
  ###
end#png


def self.dxf()
  ###
  dxffolder=File.join(File.dirname(@model.path), @model.title+"-Slicer5-DXFs-"+@rnd).tr("\\","/")
  Dir.mkdir(dxffolder) if not File.exist?(dxffolder)
  @defns=@model.definitions
  @ents=@model.entities
  @layers=@model.layers
  @view=@model.active_view
  cam=@view.camera
  @cam=cam
  @came=cam.eye
  @camt=cam.target
  @camu=cam.up
  @camp=cam.perspective?
  ###
  names=[]
  msg=@msg
  @defns.each{|defn|
    next unless defn.name=~/^#{@rnd}[-][XYZ][0-9][0-9][0-9]/
    names << defn.name if defn.instances[1] ### only export used panels
    msg=msg+"."
    Sketchup.status_text=msg
  }
  names.sort!
  return nil if not names[0]
  ###
  @model.commit_operation unless @sep ### as it's nested 
  @dxf=true
  ###
  Sketchup.status_text=@msg+"DXFs "
  ###
  camtop=Sketchup::Camera.new([0,0,1], [0,0,0], Y_AXIS)
  camtop.perspective=false
  @view.camera=camtop
  @view.refresh
  dxfs=[]
  ###
  names.each{|name| ####################################################
    ###
	@model.start_operation("Slicer5"+@ex+":DXFs", true)
	###
	defn=@defns[name]
    ###
    dxf=File.join(dxffolder, name+'.dxf').tr("\\","/")
    ###
    Sketchup.status_text=(@msg+"DXF "+name+'.dxf')
    ### sort out 2D layers
    layerE=@layers.add("EXTL")
    layerI=@layers.add("INTL")
    layerT=@layers.add("TEXT")
    ###
    ### erase ALL
    @ents.erase_entities(@ents.to_a)
    ### add defn
    tr=Geom::Transformation.new()
    inst=@ents.add_instance(defn, tr)
    ###
    ### use Slice geom [but leave Text as group]
	  ### relocate Text at z=0
	  txt=nil
	  defn.entities.each{|e|
        next unless e.class==Sketchup::Group and e.name=="TEXT"
        txt=e
      }
	  pt=txt.bounds.min
	  pt.z=0
	  tr=Geom::Transformation.translation(txt.bounds.min.vector_to(pt))
	  txt.transform!(tr)
	  ### geom - keep only bottom face
      togos=[]
      defn.entities.each{|e|
        togos << e if e.class==Sketchup::Edge and e.bounds.center.z!=0
      }
      togos.uniq!
      defn.entities.erase_entities(togos)
	  ###
	  faces=[]
	  defn.entities.to_a.each{|e|
	    faces << e if e.class==Sketchup::Face
	  }
	 faces.each{|face|
      ### EXTL outer loop
	  oloop=face.outer_loop
	  oedges=oloop.edges
	  overts=oloop.vertices
	  overts << overts[0]
	  opts=[]
	  overts.each{|v|opts << v.position}
	  ### INTL inner loops [if any]
	  loops=face.loops-[face.outer_loop]
	  iverts=[]
	  pdepths=[]
	  pverts=[]
	  circs=[]
	  ###
	  loops.each{|loop|iverts << loop.vertices+[loop.vertices[0]]}
	  ### make outer loop from opts
	  edges=@ents.add_curve(opts)
	  edges.each{|ed|ed.layer=layerE}
	  ###
	  ### make inner loops
	  ipts=[]
	  iverts.each{|vs|
	    pts=[]
	    vs.each{|v|pts << v.position}
		ipts << pts
      }
	  ### process inners
	  ipts.each{|pts|
	    edges=@ents.add_curve(pts)
	    edges.each{|ed|ed.layer=layerI}
	  }
	 }#faces
     ### sort out Text ===========================
	 tt=txt.transformation
	 int=@ents.add_instance(txt.entities.parent, tt)
	 txt.erase!
	 ### polyline text's edges
	 tents=int.entities
	 edges=tents.to_a
	 cedges=[]
	 edges.each{|e|cedges << e.all_connected}
	 cedges.each{|eds|
	   next if not eds[0].valid? or eds[0].curve
	   vs=[]
	   eds.each{|e|vs<<e.vertices}
	   vs.flatten!
	   vs.uniq!
	   vs << vs[0]
	   pts=[]
	   vs.each{|v|pts << v.position}
	   tents.erase_entities(eds)
	   tents.add_curve(pts)
	 }
	 tents.each{|e|e.layer=layerT}
	 int.explode
	 ###
    ### adjust so slice's bb.min is at origin [0,0,0]
    min=inst.bounds.min
	vec=min.vector_to(ORIGIN)
    trv=Geom::Transformation.translation(vec)
	inst.erase!
	@ents.transform_entities(trv,@ents.to_a)
	### flatten!
	pt0=@ents[0].start.position
	pt1=pt0.clone
	pt1.z=0
	vvv=pt0.vector_to(pt1)
	tvv=Geom::Transformation.translation(vvv)
	@ents.transform_entities(tvv, @ents.to_a)
    ###
    ### purge all unused compos and layers
    @defns.purge_unused
    @layers.purge_unused
    ###
    ### zoom to plan/top and extents
    @view.zoom_extents
    @view.refresh
    ###
    @model.export(dxf)
	###
	dxfs << dxf
    ###
    begin
      @model.abort_operation ### undo so back as it was !!!
    rescue Exception => e
      puts e
    end
    ###
  } ####################################################################
  ###
  dxfs.each{|dxf|self.recolorlayers(dxf)}
  ###
  ### reset camera as it was
  @view.camera.perspective=@camp
  camo=Sketchup::Camera.new(@came, @camt, @camu)
  @view.camera=camo
  ###
  Sketchup.status_text=("Slicer5"+@ex+": DXF Files Completed.")
  UI.messagebox("Slicer5"+@ex+":\n\n"+names.length.to_s+" DXF Files Exported into\n\n"+dxffolder)
  #UI.openURL('file:///'+dxffolder)
  ###
end#dxf

def self.recolorlayers(dxf)
=begin
0 		7 white/black
EXTL	2 yellow
INTL	3 green
TEXT	1 red
=end
  text=IO.read(dxf)
  ### make DOS > MAC; CR+LF > LF; return later
  pc=false
  pc=true if [PLATFORM].grep(/mswin/)==[PLATFORM]
  unless pc
    text.gsub!(/\r\n/,"\n")
  end
  ###
  bits=text.split("TABLE\n  2\nLAYER\n")
  start=bits[0]+"TABLE\n  2\nLAYER\n"
  rest=bits[1]
  table=rest.split("ENDTAB\n")[0]
  remainder=rest.split("ENDTAB\n")[1..-1].join("ENDTAB\n")
  layers=table.split("LAYER\n")
  dxfout=dxf ###+".DXF" ###
  file=File.open(dxfout,"w")
  file.puts(start)
  file.puts(layers[0])
  layers[1..-1].compact.each{|layer|
    next if not layer
    layer="LAYER\n"+layer
    parts=layer.split("  2\n")
	next if not parts[1]
	parts[0]=parts[0]+"  2\n"
	name=parts[1].split("\n")[0]
	case name
	when "0"
	  parts[1]=parts[1].gsub(/420\n[ 0-9]+\n/,"").gsub(/ 62\n[ 0-9]+\n/," 62\n     7\n")
	when "EXTL"
	  parts[1]=parts[1].gsub(/420\n[ 0-9]+\n/,"").gsub(/ 62\n[ 0-9]+\n/," 62\n     2\n")
	when "INTL"
	  parts[1]=parts[1].gsub(/420\n[ 0-9]+\n/,"").gsub(/ 62\n[ 0-9]+\n/," 62\n     3\n")
	when "TEXT"
	  parts[1]=parts[1].gsub(/420\n[ 0-9]+\n/,"").gsub(/ 62\n[ 0-9]+\n/," 62\n     1\n")
	else ### ???
	  parts[1]=parts[1].gsub(/420\n[ 0-9]+\n/,"").gsub(/ 62\n[ 0-9]+\n/," 62\n     7\n")
	end
	file.puts(parts[0])
	file.puts(parts[1])
  }
  file.puts("ENDTAB\n")
  file.puts(remainder)
  file.close
  ### return MAC > DOS;  LF > CR+LF 
  unless pc
    text=IO.read(dxf)
    text.gsub!(/\n/,"\r\n")
	file=File.open(dxfout,"w")
    file.puts(text)
	file.close
  end
  ###
end#recolorlayers


end#Slicer class

##########
end#TIG module
##########
